###################################################################################################
###################################################################################################
###################################################################################################
#
# GSTiCluster (Group Structured Tight iCluster)
# Version : 1.1.0
# Authors : SungHwan Kim, Yongseok Park and George C. Tseng
# latest update : 01/06/2015
#
###################################################################################################
###################################################################################################
###################################################################################################

GSTiCluster <- function(datasets, k, Best_lam = NULL, max.iter=100, is.scale=FALSE, f.module, is.tight=FALSE, lambda_1 = NULL) {
  # Description 
  #     Fit a group structured tight integrative clustering
  # Usage
  #     S = GSTiCluster(datasets, k, Best_lam, max.iter=100, is.scale=FALSE, f.module)
  # Input
  #     X = Input variable matrix (integrated matrix of multiple datasets)
  #     k = # of integrative clustering
  #     Best_lam = Regularization parameter
  #     lambda_1 = Regularization parameter for tight clustering
  #     f.module = Input feature module matrix (Row: feature module index / Col: feature index)
  #     is.tight = logical value indicating whether tight clustering is conducted
  #     max.iter = # of maximum iteration of EM
  # Output
  #     model = a list consists of fit, Beta, expZ
  #        model$Beta = Estimated beta coefficients (n by k-1)
  #        model$expZ = Estimated average values of latent variables (n by k-1)
  #        model$cluster = Group structured clustering labels of samples
  #        model$tight.cluster = Group structured tight clustering labels of samples
  #        model$is.sample.penalized = logical values of whether samples are excluded
  
  # Data Initialization
  datasets <- t(datasets)
  n = dim(datasets)[1]
  sum.p = dim(datasets)[2]
  
  stacked <- datasets
  if(is.scale){
    stacked = scale(stacked, center=TRUE, scale=TRUE)    
  } else{
    stacked = scale(stacked, center=TRUE, scale=FALSE)    
  }
  
  # Initialization of Beta (=W)
  svd.x = svd(stacked)
  v = svd.x$v[, 1:(k - 1), drop = F]
  d = svd.x$d^2
  sigma2 = sum(svd.x$d[k:n])/(n - k + 1)
  cc = sqrt(d[1:(k - 1)] - sigma2)
  n.cc = length(cc)
  W = v %*% diag(cc, nrow = n.cc, ncol = n.cc)
  
  
  # Initialization of inverse Sigma
  PSI = rep(sigma2, sum.p)
  inv.phi <- solve(diag(PSI))
  
  sigma = W %*% t(W)
  diag.idx = 1 + 0:(sum.p - 1) * (sum.p + 1)
  sigma[diag.idx] = sigma[diag.idx] + PSI
  inv.sigma = solve(sigma)
  
  # Initialization of Z
  expZ = matrix(0, nrow = k-1, ncol = n)
  
  # Initialization of parameters
  iter=1
  conv.rate=1
  save.conv.rate=NULL
  cat(paste("# of cluster =",k,":" ,"\n",sep=" "))
  epsilon = 1e-3
  
  while((iter < max.iter) & (conv.rate > epsilon)){
    
    ##########################
    #  E-step
    ##########################
    expZ.old = expZ
    expZ = t(W) %*% inv.sigma %*% t(stacked)
    if(sum(apply(expZ,1, function(x) all(x==0))) == dim(expZ)[1]){
      break
    }

    expZ <- as.matrix(t(scale(t(expZ))))
    varZ = diag(1,nrow=k-1,ncol=k-1) - t(W)%*%inv.sigma%*%W
    expZZ = varZ + expZ%*%t(expZ)  
    
    ########################
    #  M-step
    ########################
    W.old = W
    data.Y <- stacked
    data.X <- t(expZ)
    data.XX <- expZZ
    
    Object <- OverLasso_pre_grad(data.X, data.Y, data.XX, inv.phi, W, T = f.module, is.balanced = TRUE, is.weight.g=FALSE, prop.g=1)  ## is.weight.g=FALSE means we don't manipulate the weight
    X = Object$X
    Y = Object$Y
    XX = Object$XX
    XY = Object$XY
    C = Object$C
    g_idx = Object$g_idx
    TauNorm = Object$TauNorm
    L1 = Object$L1
    L1_Sig = Object$L1_Sig
    XSigX = Object$XSigX
    XSigY = Object$XSigY
    SigBXX = Object$SigBXX
    option.maxiter = 1000;

    # To circumvent over-regulation at the first iteration.
    if(iter==1){
      option.threshold = 0
      best_lam = Best_lam*0.005
    } else {
      option.threshold = 1e-3
      best_lam = Best_lam
    }
    option.tol = 1e-3
    mu = 0.01
    
    
    ## Sparse overlapping group lasso
    L = L1_Sig + best_lam^2*TauNorm/mu  
    Result <- OverLasso_accgrad(Y, X, best_lam, T = f.module, data.XX, t(XSigY), inv.phi, C, g_idx, L, mu, option.maxiter, option.tol, option.threshold, option.verbose=FALSE, is.accelation=TRUE)
    W <- t(Result$Beta)
    a = diag( (t(stacked  - t(expZ)%*%t(W)) %*% (stacked  - t(expZ)%*%t(W)) ) )
    PSI = a/n 
    inv.phi <- solve(diag(PSI))
    sigma = W %*% t(W)   
    sigma[diag.idx] = sigma[diag.idx] + PSI
    inv.sigma = solve(sigma)
    
    # Convergence rate
    conv.rate = max(abs(W - W.old))
    save.conv.rate[iter] = conv.rate
    iter=iter+1
    cat("Iteration:", iter,"\n")
  }
  
  cls <- pam(t(expZ), k=k)
  clustering <- cls$clustering
  
    
  if(is.tight){
    tight.result <- TightiCluster(stacked, W, k, lambda_1 = lambda_1)
    return( list(Beta = W, Z = expZ, cluster=clustering, tight.expZ = tight.result$tight.expZ, is.sample.penalized = tight.result$is.sample.penalized, tight.cluster = tight.result$tight.cluster ))
    } else{
    return( list(Beta = W, Z = expZ, cluster=clustering) )  
  }
}



OverLasso_pre_grad <- function(X_orig, Y_orig, data.XX, inv.phi, Beta, T, is.balanced=FALSE, is.weight.g=FALSE, prop.g=1){
  N = dim(X_orig)[1]
  V = dim(T)[1]
  K = dim(T)[2]
  
  Y <- Y_orig
  X <- X_orig
  
  sum_col_T <- colSums(t(T))  
  SV = sum(sum_col_T)
  csum = cumsum(sum_col_T)
  g_idx = cbind(c(1, csum[-length(csum)]+1), csum , sum_col_T)  
  J = rep(0, SV)
  W = rep(0, SV)
  
  for( v in 1:V){
    J[g_idx[v,1]:g_idx[v,2]] = rbind(1:dim(T)[2], T[v,])[1,T[v,]==1] 
    W[g_idx[v,1]:g_idx[v,2]] = 1  
  }
  
  .ind <- cbind(1:SV, J)
  .tmpT <- matrix(0, SV, K)
  for(ii in 1:dim(.ind)[1]) .tmpT[.ind[ii,1], .ind[ii,2]] <- W[ii]
  C <- .tmpT
  
  inv_colSum <-  1/colSums(C)
  inv_colSum[inv_colSum == Inf] = 0
  
  if(is.balanced){
    C <- t(t(C) * inv_colSum) # To make sure the sum of values equal to 1.
    
    if(is.weight.g){
      C <- C*prop.g
    }
    
  } else {
    matrix(as.numeric(C>0),nrow(C),ncol(C))
  }
  
  XX = data.XX
  XY = t(X) %*% Y
  
  SigBXX = inv.phi  %*% Beta %*% data.XX
  XSigX =  inv.phi  %*% matrix(1, dim(Y)[2], dim(X)[2]) %*% data.XX
  XSigY =  inv.phi  %*% t(Y) %*% X
  
  inv_w <-  1/colSums(T)
  inv_w[inv_w == Inf] = 0
  
  TauNorm <- t(t(T) * inv_w)
  TauNorm = max(colSums(TauNorm^2))
  
  L1 = eigen(XX)$values[1]
  
  tmp <- foreach(i = 1:dim(Y)[2],.combine=c) %do% {
    rep(max(diag(data.XX),0.01)*diag(inv.phi)[i], dim(X)[2])
  }
  L1_Sig <- max(tmp)
  
  return( list(X=X, Y=Y, XX=XX, XY=XY, C=C, g_idx=g_idx, TauNorm=TauNorm, L1=L1, L1_Sig=L1_Sig, XSigX=XSigX , XSigY=XSigY , SigBXX=SigBXX) )
  
}






OverLasso_accgrad <- function(Y, X, lambda, T, data.XX, XSigY, inv.phi, C, g_idx, L, mu, option.maxiter,
                              option.tol, option.threshold, option.verbose=FALSE, is.accelation=TRUE  ){
  
  J = dim(X)[2]
  K = dim(T)[2]
  maxiter = option.maxiter
  tol = option.tol
  threshold = option.threshold
  verbose = option.verbose
  obj=rep(0, maxiter)
  time=rep(0, maxiter)
  C = C*lambda;
  bw = matrix(0,J,K)
  bx = bw    
  theta=1
  conv.rate = 1
  ptm <- proc.time()
  
  
  for ( .iter in 1:maxiter){
    
    R = shrink( C %*% t(bw)/mu, g_idx)
    grad_bw = (data.XX %*% bw %*% inv.phi) - XSigY + (t(R) %*% C) 
    
    bv = bw - (1/L)*grad_bw 
    
    bx_new = sign(bv) * ((abs(bv) - (lambda) /L) * ( (abs(bv)- (lambda)/L) > 0)) 
    
    obj[.iter] = sum(   diag(   (Y-(X%*%bx_new))%*% inv.phi %*%t(Y-(X%*%bx_new)) )   )/2 + cal2norm(C %*% t(bx_new), g_idx)
        
    if(is.accelation){
      theta_new = 2 / (.iter+2)  # % because iter starts with 1
      bw = bx_new + (1-theta) / theta*theta_new*(bx_new-bx) 
      time[.iter] =  (proc.time() - ptm)[2]
      if(option.verbose) cat(paste('Iter=', .iter, "Obj=" , obj[.iter] ))
      
      ## Updating step
      theta = theta_new
      bx.old = bx
      bx = round(bx_new,4)
      
    }else{
      if(option.verbose) cat(paste('Iter=', .iter, "Obj=" , obj[.iter] ))
      time[.iter] =  (proc.time() - ptm)[2]
      bw  =  bx_new
      bx.old = bx
      bx = round(bx_new,4)
    }
    conv.rate <- max(abs(bx - bx.old))
    if(.iter > 2 && ( (abs(obj[.iter] - obj[.iter-1])/abs(obj[.iter-1])) < tol)) {break}
    .iter <- .iter + 1
  }
  
  Beta <- bx*(!(abs(bx) < threshold))
  obj = obj[1:.iter]
  time = time[1:.iter]
  return( list(Beta=Beta, obj=obj, time=time) )
}


shrink <- function(A, g_idx){
  V = dim(g_idx)[1]
  R = matrix(0, dim(A)[1], dim(A)[2])
  for( v in 1:V){
    idx = g_idx[v,1]:g_idx[v,2]
    
    if(dim(A)[2]==1){
      gnorm = sqrt(sum(A[idx, ]^2))  
    }else{
      if(dim(A)[1]==1){
        gnorm = sqrt((A[idx, ]^2))    
      }else{
        
        if( length(idx)==1){
          gnorm = sqrt((A[idx, ]^2))
        } else {
          gnorm = sqrt(colSums(A[idx, ]^2))    
        }
        
      }
    }
    
    gnorm[gnorm<1] = 1
    R[idx, ] = A[idx,] / t(matrix(rep(gnorm), length(gnorm) , g_idx[v,3]))
  }
  return(R)
}

cal2norm <- function(A, g_idx){
  V = dim(g_idx)[1]
  s = 0
  for(v in 1:V){
    idx = g_idx[v,1]:g_idx[v,2]
    if(dim(A)[2]==1){
      gnorm = sqrt(sum(A[idx, ]^2))  
    }else{
      
      if(dim(A)[1]==1){
        gnorm = sqrt((A[idx, ]^2))    
      }else{
        
        if( length(idx)==1){
          gnorm = sqrt((A[idx, ]^2))
        } else {
          gnorm = sqrt(colSums(A[idx, ]^2))    
        }
      }
    }
    
    s = s + sum(gnorm)
  }
  return(s)
}



TightiCluster <- function(stacked, Beta, k, lambda_1 ){
  W <- Beta[ ,!apply(Beta,2, function(x) all(x==0))]
  expZ <- t(solve(t(W)%*%W)%*%t(W)%*%t(stacked))
  expZ = soft(expZ, lambda_1)
  
  pen.ind <- foreach( i = 1:dim(expZ)[1], .combine=c ) %do% { !all(expZ[i,]==0)  }
  expZ <- t(expZ[pen.ind,])
  cls <- pam(t(expZ), k=k)$clustering
  
  return(list(tight.expZ = expZ, is.sample.penalized = pen.ind, tight.cluster = cls))
}

soft <- function (a, para){
  b <- sort(abs(a))
  b <- abs(a) - para
  b <- (b + abs(b))/2 
  b <- sign(a) * b
  b
}

